@@ -13,8 +13,8 @@ import square.schemas
 import yaml
 import yaml.scanner
 from square.dtypes import (
-    Config, Filepath, GroupBy, K8sConfig, LocalManifestLists, LocalManifests,
-    MetaManifest, Selectors, ServerManifests,
+    Config, Filepath, GroupBy, K8sConfig, LocalManifestLists, MetaManifest,
+    Selectors, ServerManifests,
 )
 
 # Convenience: global logger instance to avoid repetitive code.
@@ -62,6 +62,11 @@ def select(manifest: dict, selectors: Selectors) -> bool:
         logit.error(f"BUG: selector must specify a `kind`: {selectors}")
         return False
 
+    # Split all labels into tuples: "foo=bar" -> ("foo", "bar").
+    # Sanity check: hard abort if the labels do not split.
+    label_selectors = {tuple(_.split("=")) for _ in selectors.labels}
+    assert all([len(_) == 2 for _ in label_selectors])
+
     # Unpack the resource's kind and labels.
     kind = manifest.get("kind", None)
     labels = manifest.get("metadata", {}).get("labels", {})
@@ -108,9 +113,9 @@ def select(manifest: dict, selectors: Selectors) -> bool:
     # labels or not.
     labels = {(k, v) for k, v in labels.items()}
 
-    # The resource must match the label selectors, if we have any such selectors.
-    if selectors.labels:
-        if not selectors.labels.issubset(labels):
+    # The resource must match the label selectors unless we have no such selectors.
+    if label_selectors:
+        if not label_selectors.issubset(labels):
             logit.debug(f"Labels {labels} do not match selector {selectors.labels}")
             return False
 
@@ -414,9 +419,9 @@ def filename_for_manifest(
 
 def diff(config: Config,
          k8sconfig: K8sConfig,
-         local: LocalManifests,
-         server: ServerManifests) -> Tuple[str, bool]:
-    """Return the human readable diff between the `local` and `server`.
+         local: dict,
+         server: dict) -> Tuple[str, bool]:
+    """Return the human readable diff between the `local` and `server` manifest.
 
     The diff shows the necessary changes to transition the `server` manifest
     into the state of the `local` manifest.
@@ -434,17 +439,9 @@ def diff(config: Config,
         produce it.
 
     """
-    # Clean up the input manifests because we do not want to diff, for instance,
-    # the `status` fields.
-    srv, _, err1 = strip(k8sconfig, server, config.filters)
-    loc, _, err2 = strip(k8sconfig, local, config.filters)
-    if err1 or err2:
-        return ("", True)
-
-    # Undo the DotDicts. This is a precaution because the YAML parser can
-    # otherwise not dump the manifests.
-    srv = square.dotdict.undo(srv)
-    loc = square.dotdict.undo(loc)
+    # Precaution: undo the DotDicts to ensure the YAML parse will accept them.
+    srv = square.dotdict.undo(server)
+    loc = square.dotdict.undo(local)
     srv_lines = yaml.dump(srv, default_flow_style=False).splitlines()
     loc_lines = yaml.dump(loc, default_flow_style=False).splitlines()
 
@@ -553,7 +550,7 @@ def align_serviceaccount(
     namespace.
 
     This makes it difficult to manage service accounts with Square because the
-    token is not known in advance. Once would have to
+    token is not known in advance. One would have to
 
         square apply; square plan; square get serviceaccount
 
@@ -562,8 +559,8 @@ def align_serviceaccount(
 
     To avoid this problem, this function will read the token secret that K8s
     added (contained in `server_manifest`) and insert it into the
-    `local_manifest`. This will ensure that that Square can create a plan that
-    will not touch the token secret.
+    `local_manifest`. This will ensure that Square creates a plan that
+    leaves the token secret alone.
 
     Inputs:
         local_manifests: manifests from local files that the plan will use.
@@ -953,7 +950,7 @@ def download(config: Config, k8sconfig: K8sConfig) -> Tuple[ServerManifests, boo
                 manifests, err = unpack_list(manifest_list, config.selectors)
                 assert not err and manifests is not None
 
-                # Drop all manifest fields except "apiVersion", "metadata" and "spec".
+                # Strip off the fields defined in `config.filters`.
                 ret = {k: strip(k8sconfig, man, config.filters)
                        for k, man in manifests.items()}
 
