@@ -1,5 +1,8 @@
+import copy
 import json
 import logging
+import re
+from collections import Counter
 from types import SimpleNamespace
 from typing import Collection, Optional, Set, Tuple
 
@@ -61,39 +64,63 @@ def load_config(fname: Filepath) -> Tuple[Config, bool]:
     # translate verbatim.
     cfg.folder = fname.parent.absolute() / cfg.folder
     cfg.selectors.kinds = set(cfg.selectors.kinds)
-    cfg.selectors.labels = {(k, v) for k, v in cfg.selectors.labels}
     return cfg, False
 
 
+def translate_resource_kinds(cfg: Config, k8sconfig: K8sConfig) -> Config:
+    """Convert `cfg.Selectors.kind` and `cfg.priorities` to their canonical names.
+
+    Example: "svc" -> "Service" or "ns" -> "Namespace".
+
+    Silently ignore unknown resource kinds. This is necessary because the user
+    may have specified a custom resource that does not (yet) exist. Since we cannot
+    distinguish them from typos, we allow unknown resource kinds but
+    get/plan/apply cycles will ignore them.
+
+    """
+    # Avoid side effects to the original `cfg`.
+    cfg = copy.deepcopy(cfg)
+
+    # Translate the possibly colloquial names to their canonical K8s names. Use
+    # the original name if we cannot find the canonical name.
+    kinds = {k8sconfig.short2kind.get(_.lower(), _) for _ in cfg.selectors.kinds}
+    priorities = [k8sconfig.short2kind.get(_.lower(), _) for _ in cfg.priorities]
+
+    # Return a copy of `cfg` with the translated resource kinds.
+    cfg.selectors.kinds.clear()
+    cfg.selectors.kinds.update(kinds)
+    cfg.priorities.clear()
+    cfg.priorities.extend(priorities)
+    return cfg
+
+
 def make_patch(
         config: Config,
         k8sconfig: K8sConfig,
-        local: ServerManifests,
-        server: ServerManifests) -> Tuple[JsonPatch, bool]:
+        local: dict,
+        server: dict) -> Tuple[JsonPatch, bool]:
     """Return JSON patch to transition `server` to `local`.
 
     Inputs:
         config: Square configuration.
         k8sconfig: K8sConfig
-        local: LocalManifests
-            Usually the dictionary keys returned by `load_manifest`.
-        server: ServerManifests
-            Usually the dictionary keys returned by `manio.download`.
+        local: dict
+            Usually on fo the dict manifests returned by `load_manifest`.
+        server: dict
+            Usually on fo the dict manifests returned by `manio.download`.
 
     Returns:
         Patch: the JSON patch and human readable diff in a `Patch` tuple.
 
     """
-    # Reduce local and server manifests to salient fields (ie apiVersion, kind,
-    # metadata and spec). Abort on error.
-    loc, _, err1 = manio.strip(k8sconfig, local, config.filters)
-    srv, _, err2 = manio.strip(k8sconfig, server, config.filters)
-    if err1 or err2 or loc is None or srv is None:
-        return (JsonPatch("", []), True)
+    # Convenience.
+    loc, srv = local, server
+    meta = manio.make_meta(local)
 
     # Log the manifest info for which we will try to compute a patch.
-    man_id = f"{loc.kind.upper()}: {loc.metadata.name}/{loc.metadata.name}"
+    man_id = f"{meta.kind}: {meta.namespace}/{meta.name}"
     logit.debug(f"Making patch for {man_id}")
+    del meta
 
     # Sanity checks: abort if the manifests do not specify the same resource.
     try:
@@ -237,14 +264,38 @@ def compile_plan(
     """
     err_resp = (DeploymentPlan(tuple(), tuple(), tuple()), True)
 
+    # Apply the filters to all local and server manifests before we compute patches.
+    server = {
+        meta: manio.strip(k8sconfig, man, config.filters)
+        for meta, man in server.items()
+    }
+    local = {
+        meta: manio.strip(k8sconfig, man, config.filters)
+        for meta, man in local.items()
+    }
+
+    # Abort if any of the manifests could not be stripped.
+    err_srv = {_[2] for _ in server.values()}
+    err_loc = {_[2] for _ in local.values()}
+    if True in err_srv or True in err_loc:
+        logit.error("Could not strip all manifests.")
+        return err_resp
+
+    # Unpack the stripped manifests (`manio.strip` returned a tuple with three
+    # entries, and here we unpack the first one because it is the stripped manifest).
+    server = {k: square.dotdict.undo(v[0]) for k, v in server.items()}
+    local = {k: square.dotdict.undo(v[0]) for k, v in local.items()}
+
     # Replace the API group of the local resource with the one K8s prefers.
     local, err = preferred_api(k8sconfig, local)
     if err:
+        logit.error("Could not determine the preferred APIs for all manifests.")
         return err_resp
 
     # Partition the set of meta manifests into create/delete/patch groups.
     plan, err = partition_manifests(local, server)
-    if err or not plan:
+    if err:
+        logit.error("Could not partition the manifests for the plan.")
         return err_resp
 
     # Sanity check: the resources to patch *must* exist in both local and
@@ -252,30 +303,35 @@ def compile_plan(
     assert set(plan.patch).issubset(set(local.keys()))
     assert set(plan.patch).issubset(set(server.keys()))
 
-    # Compile the Deltas to create the missing resources.
-    create = []
-    for delta in plan.create:
-        # We only need the resource and namespace, not its name, because that
-        # is how the POST request to create a resource works in K8s.
-        resource, err = k8s.resource(k8sconfig, delta._replace(name=""))
-        if err or not resource:
-            return err_resp
-        create.append(DeltaCreate(delta, resource.url, local[delta]))
-
-    # Compile the Deltas to delete the excess resources. Every DELETE request
-    # will have to pass along a `DeleteOptions` manifest (see below).
+    # For later: every DELETE request will have to pass along a `DeleteOptions`
+    # manifest (see below).
     del_opts = {
         "apiVersion": "v1",
         "kind": "DeleteOptions",
         "gracePeriodSeconds": 0,
         "orphanDependents": False,
     }
+
+    # Compile the Deltas to create the missing resources.
+    create = []
+    for delta in plan.create:
+        # We only need the resource and namespace, not its name, because that
+        # is how the POST request to create a resource works in K8s.
+        # Ignore the error flag because the `strip` function we used above
+        # already ensured the resource exists.
+        resource, err = k8s.resource(k8sconfig, delta._replace(name=""))
+        assert not err
+
+        # Compile the Delta and add it to the list.
+        create.append(DeltaCreate(delta, resource.url, local[delta]))
+
+    # Compile the Deltas to delete the excess resources.
     delete = []
     for meta in plan.delete:
-        # Resource URL.
+        # Resource URL. Ignore the error flag because the `strip` function
+        # above already called `k8s.resource` and would have aborted on error.
         resource, err = k8s.resource(k8sconfig, meta)
-        if err or not resource:
-            return err_resp
+        assert not err
 
         # Compile the Delta and add it to the list.
         delete.append(DeltaDelete(meta, resource.url, del_opts.copy()))
@@ -288,12 +344,14 @@ def compile_plan(
         # Compute textual diff (only useful for the user to study the diff).
         diff_str, err = manio.diff(config, k8sconfig, local[meta], server[meta])
         if err or diff_str is None:
+            logit.error(f"Could not compute the diff for <{meta}>.")
             return err_resp
 
         # Compute the JSON patch that will change the K8s state to match the
         # one in the local files.
         patch, err = make_patch(config, k8sconfig, local[meta], server[meta])
         if err or patch is None:
+            logit.error(f"Could not compute the patch for <{meta}>")
             return err_resp
 
         # Append the patch to the list of patches, unless it is empty.
@@ -468,6 +526,72 @@ def setup_logging(log_level: int) -> None:
     logit.info(f"Set log level to {level}")
 
 
+def sort_plan(cfg: Config, plan: DeploymentPlan) -> Tuple[DeploymentPlan, bool]:
+    """Return a copy of the `plan` where the entries are sorted by priority.
+
+    For example, if `cfg.priorities = ["Namespace", "Deployment"]` then
+    `plan.create` will first list all Namespace resources, then "Deployment"
+    resources, then everything else. The same applies to `plan.delete` except
+    the list will be reversed, ie the top entries are the ones missing from
+    `cfg.priorities`, followed by "Deployments", followed by "Namespaces".
+
+    """
+    # Return with an error if the entries in `cfg.priorities` are not unique.
+    if len(set(cfg.priorities)) < len(cfg.priorities):
+        duplicates = {k for k, v in Counter(cfg.priorities).items() if v > 1}
+        logit.error(f"Found duplicates in the priorities: {duplicates}")
+        return plan, True
+
+    # -------------------------------------------------------------------------
+    # The algorithm proceeds as follows: assign in `cfg.priorities` a value
+    # (resources with a higher value have a lower priority). Then it will
+    # prefix the tuples in `plan.{create,delete}` with the priority ID and sort
+    # the new list. This will ensure the all resources appear in the order
+    # defined in `cfg.priorities`. Resources with the same priority will be
+    # sorted by MetaManifest, which means sorted by namespace, the name.
+    # -------------------------------------------------------------------------
+
+    # All unknown resource kinds will have this priority, which is larger (ie
+    # less important) than all the resource kinds that are in `cfg.priorities`.
+    max_id = len(cfg.priorities)
+
+    # Assign each resource kind a number in order of priority.
+    priority_id = {name: idx for idx, name in enumerate(cfg.priorities)}
+
+    # Assign all resource kinds that exist in `plan.{create,delete}` a priority
+    # number that is larger than all other numbers in that list. This will
+    # ensure they come last when we sort.
+    missing_create = {_.meta.kind for _ in plan.create if _.meta.kind not in priority_id}
+    priority_id.update({kind: max_id for kind in missing_create})
+    missing_delete = {_.meta.kind for _ in plan.delete if _.meta.kind not in priority_id}
+    priority_id.update({kind: max_id for kind in missing_delete})
+
+    # Sort the patches by priority ID and MetaManifest.
+    create = [(priority_id[_.meta.kind], _) for _ in plan.create]
+    create.sort(key=lambda _: _[:2])
+
+    # Repeat for the patches that will delete resources but reverse the final
+    # list. This will ensure we remove resources in the reverse order in which
+    # we would create them.
+    delete = [(priority_id[_.meta.kind], _) for _ in plan.delete]
+    delete.sort(key=lambda _: _[:2])
+    delete.reverse()
+
+    # Assemble the final deployment plan and return it.
+    out = DeploymentPlan(
+        create=[_[1] for _ in create],
+        patch=list(plan.patch),
+        delete=[_[1] for _ in delete],
+    )
+    return out, False
+
+
+def valid_label(label: str) -> bool:
+    """Return `True` if `label` is K8s and Square compatible."""
+    pat = re.compile(r"^[a-z0-9][-a-z0-9_.]*=[-A-Za-z0-9_.]*[A-Za-z0-9]$")
+    return pat.match(label) is not None
+
+
 def apply_plan(cfg: Config, plan: DeploymentPlan) -> bool:
     """Update K8s resources according to the `plan`.
 
@@ -479,11 +603,23 @@ def apply_plan(cfg: Config, plan: DeploymentPlan) -> bool:
         None
 
     """
+    # Sanity check labels.
+    if not all([valid_label(_) for _ in cfg.selectors.labels]):
+        logit.error(f"Invalid labels: {cfg.selectors.labels}")
+        return True
+
     try:
+        # Sort the plan according to `cfg.priority`.
+        plan, err = sort_plan(cfg, plan)
+        assert not err
+
         # Create properly configured Requests session to talk to K8s API.
         k8sconfig, err = k8s.cluster_config(cfg.kubeconfig, cfg.kubecontext)
         assert not err
 
+        # Convert "Selectors.kinds" to their canonical names.
+        cfg = translate_resource_kinds(cfg, k8sconfig)
+
         # Create the missing resources.
         for data_c in plan.create:
             print(f"Creating {data_c.meta.kind.upper()} "
@@ -519,11 +655,19 @@ def make_plan(cfg: Config) -> Tuple[DeploymentPlan, bool]:
         Deployment plan.
 
     """
+    # Sanity check labels.
+    if not all([valid_label(_) for _ in cfg.selectors.labels]):
+        logit.error(f"Invalid labels: {cfg.selectors.labels}")
+        return DeploymentPlan(tuple(), tuple(), tuple()), True
+
     try:
         # Create properly configured Requests session to talk to K8s API.
         k8sconfig, err = k8s.cluster_config(cfg.kubeconfig, cfg.kubecontext)
         assert not err
 
+        # Convert "Selectors.kinds" to their canonical names.
+        cfg = translate_resource_kinds(cfg, k8sconfig)
+
         # Load manifests from local files.
         local, _, err = manio.load(cfg.folder, cfg.selectors)
         assert not err
@@ -548,18 +692,26 @@ def make_plan(cfg: Config) -> Tuple[DeploymentPlan, bool]:
 
 def get_resources(cfg: Config) -> bool:
     """Download all K8s manifests and merge them into local files."""
+    # Sanity check labels.
+    if not all([valid_label(_) for _ in cfg.selectors.labels]):
+        logit.error(f"Invalid labels: {cfg.selectors.labels}")
+        return True
+
     try:
         # Create properly configured Requests session to talk to K8s API.
         k8sconfig, err = k8s.cluster_config(cfg.kubeconfig, cfg.kubecontext)
         assert not err
 
+        # Convert "Selectors.kinds" to their canonical names.
+        cfg = translate_resource_kinds(cfg, k8sconfig)
+
         # Use a wildcard Selector to ensure `manio.load` will read _all_ local
         # manifests. This will allow `manio.sync` to modify the ones specified by
         # the `selector` argument only, delete all the local manifests, and then
         # write the new ones. This logic will ensure we never have stale manifests
         # (see `manio.save_files` for details and how `manio.save`, which we call
         # at the end of this function, uses it).
-        load_selectors = Selectors(kinds=k8sconfig.kinds, labels=set(), namespaces=[])
+        load_selectors = Selectors(kinds=k8sconfig.kinds, labels=[], namespaces=[])
 
         # Load manifests from local files.
         _, local, err = manio.load(cfg.folder, load_selectors)
